#include "iome.h"

#define METHOD_NAME "method.zmdbosh"
#define CIN_FILE    "zmdbosh.cin"

/***************************************************************\
                        Key Map Functions
\***************************************************************/

typedef struct {
  int    key;
  char * name;
} KeyMAP;

typedef struct {
  int      Size;
  KeyMAP * MAP;
} KeyCODE;

int AddKeyMap(KeyCODE * KC,int key,char *name)
{
  KC->MAP = (KeyMAP*)realloc(KC->MAP,sizeof(KeyMAP)*(KC->Size+1));
  KC->MAP[KC->Size].key  = key;
  KC->MAP[KC->Size].name = strdup(name);
  KC->Size++;
  return KC->Size;
}

char * WhichKEY(KeyCODE * KC,int c)
{
  int n;
  for (n=0;n<KC->Size;n++) 
    if (KC->MAP[n].key==c) return KC->MAP[n].name;
  return NULL;
}
/***************************************************************\
                              Table
\***************************************************************/
typedef struct {
  char * KEY;
  char * WORD;
} KeyToWord;

#define MAX_MATCH 2048
typedef struct {
  int         Size;
  int         Matches;
  int         SameKey;
  int         Now;
  int         NextAvail;
  int         MATCH[MAX_MATCH];
  KeyToWord * KTW;
} KeyToWordMap;

int AddKeyToWordMap(KeyToWordMap *KTW,char *key,char *word)
{
  KTW->KTW = (KeyToWord*)realloc(KTW->KTW,sizeof(KeyToWord)*(KTW->Size+1));
  KTW->KTW[KTW->Size].KEY  = strdup(key);
  KTW->KTW[KTW->Size].WORD = strdup(word);
  KTW->Size++;
  return KTW->Size;
}

int SearchKEY(KeyToWordMap *KTW,char *key,int start,int size)
{
  for (;start<KTW->Size;start++)
    if (strncmp(KTW->KTW[start].KEY,key,size)==0) return start;
  return -1;
}

int BuildMatchList(KeyToWordMap *KTW,char *key,int size)
{
  int id=-1;
  KTW->Matches = 0;
  KTW->Now     = 0;
  KTW->SameKey = 0;
  do {
    id = SearchKEY(KTW,key,id+1,size);
    if (id>=0) {
      KTW->MATCH[KTW->Matches] = id;
      KTW->Matches++;
    }
  } while (id>=0);
  /* Check if they are the same key */
  if (KTW->Matches==1) KTW->SameKey = 1; else
  if (KTW->Matches>1) {
    char * WORD = KTW->KTW[KTW->MATCH[0]].KEY;
    KTW->SameKey = 1;
    id = 1;
    while (KTW->SameKey&&id<KTW->Matches) {
      if (strcmp(WORD,KTW->KTW[KTW->MATCH[id]].KEY)!=0) KTW->SameKey = 0;
      id++;
    }
  }
  return KTW->Matches;
}

/***************************************************************/
typedef struct {
  int          Size;
  int          Nomatch;
  char         KEY[64];
  KeyCODE      KC;
  KeyToWordMap KTW;
} INMD;

int LoadCIN(char * FileName,INMD *inmd)
{
  FILE * fp;
  char buff[IOME_MAX_BUFF];
  char *w1;
  char *w2;
  int State = 0;
  int Section = -1;

  fp = fopen(FileName,"rt");
  if (fp==NULL) return 0;
  do {
    memset(buff,0,IOME_MAX_BUFF);
    if (fgets(buff,IOME_MAX_BUFF,fp)==NULL) State=1;
    if (!State&&strlen(buff)>0&&buff[0]!='#') {
      /* Begin to separate two items */
      w1 = buff;
      while (*w1&&*w1!=' '&&*w1!=9) w1++;
      *w1 = 0; w1++;
      while (*w1&&(*w1==' '||*w1==9)) w1++;
      w2 = w1; w2++;
      while (*w2&&*w2!=' '&&*w2!=9&&*w2!=10&&*w2!=13) w2++;
      *w2 = 0;
      if (buff[0]=='%') {
        if (strcmp(buff,"%ename")==0) {
          /* Do nothing */
        } else
        if (strcmp(buff,"%prompt")==0) {
          /* Do nothing */
        } else
        if (strcmp(buff,"%selkey")==0) {
          /* Do something */
        } else
        if (strcmp(buff,"%dupsel")==0) {
          /* What is that? */
        } else
        if (strcmp(buff,"%keyname")==0) {
          if (strcmp(w1,"begin")==0) Section = 6; else
          if (strcmp(w1,"end")==0)   Section = -1;
        } else
        if (strcmp(buff,"%quick")==0) {
          if (strcmp(w1,"begin")==0) Section = 7; else
          if (strcmp(w1,"end")==0)   Section = -1;
        } else
        if (strcmp(buff,"%sel1st")==0) {
          if (strcmp(w1,"begin")==0) Section = 8; else
          if (strcmp(w1,"end")==0)   Section = 9;
        }
      } else {
        /* Keep Reading */
        switch (Section) {
          case 6: /* Read Key Map */
            AddKeyMap(&inmd->KC,(int)buff[0],w1);
          break;
          case 7: /* Read Quick */
            
          break;
          case 8: /* Read Select First */

          break;
          case 9: /* Read Table */
            AddKeyToWordMap(&inmd->KTW,buff,w1);
          break;
        }
      }
    }
  } while (!State);
  fclose(fp);

  return 1;
}


char keyrow[]="qwertyuiop"
              "asdfghjkl;"
              "zxcvbnm,./";

unsigned long CONVT(char *s)
{
  unsigned long kk;
  memcpy(&kk,s,4);
  return kk;
}

int imeRegister(iomeSysInfo *INFO,iomeRegistry *REG)
{
  char *Accept_System[]={
    "x",
    "tty",
    NULL
  };
  char **IME = Accept_System;
  int n = 0;

  while (n>=0&&IME[n]) {
    if (strcasecmp(INFO->system,IME[n])==0) n=-2;
    n++;
  }
  if (n>=0) return 0;
  REG->encoding      = "big5";
  REG->class         = "ime";
  REG->accept_system = Accept_System;
  return 1;
}

int imeInit(iomeConfig *CONF,iomeRecord *RECORD)
{
  INMD * inmd;
  char * CIN;
  int n;

  inmd = (INMD*)malloc(sizeof(INMD));
  memset(inmd,0,sizeof(INMD));
  
  RECORD->MethodName = iomeWhichItem(CONF,METHOD_NAME);
  RECORD->FilterName = NULL;
  RECORD->Encoding   = "big5";
  RECORD->ListArea   = (char *)malloc(64); RECORD->ListArea[0]=0;
  RECORD->KeyArea    = (char *)malloc(64); RECORD->KeyArea[0]=0;
  RECORD->Value      = (char *)malloc(12); RECORD->Value[0]=0;
  RECORD->Owner      = inmd;

  CIN = iomeWhichItem(CONF,CIN_FILE);
  if (CIN!=NULL&&LoadCIN(CIN,inmd)) {
    /* Load Success */
    return 1;
  } else
    printf(
      "Can not load cin file: %s\n",
      (CIN==NULL) ? "No file specified":
                    CIN
    );

  return 0;
}

int imePrepare(iomeRecord *RECORD)
{
  INMD *inmd = RECORD->Owner;
  inmd->Size = 0;
  memset(inmd->KEY,0,64);
  memset(RECORD->ListArea,0,64);
  memset(RECORD->KeyArea,0,64);
  memset(RECORD->Value,0,12);
  return 1;
}

int PrintWord(INMD *inmd,int id,char *buff)
{
  if (strlen(inmd->KTW.KTW[id].KEY)==inmd->Size) {
    strcpy(buff,inmd->KTW.KTW[id].WORD);
  } else {
    sprintf(
      buff,"%c%s",
      inmd->KTW.KTW[id].KEY[inmd->Size],
      inmd->KTW.KTW[id].WORD
    );
  }
  return 1;
}

int ShowListArea(iomeRecord *RECORD,INMD *inmd,int isList)
{
  int id = inmd->KTW.Now;
  int c,n=1;
  int len;
  char buff[IOME_MAX_BUFF];

  RECORD->ListArea[0] = 0;
  
  do {
    len  = strlen(RECORD->ListArea)+4;
    len += strlen(inmd->KTW.KTW[inmd->KTW.MATCH[id]].WORD);
    if (len<RECORD->MaxListArea) {
      c = inmd->KTW.KTW[inmd->KTW.MATCH[id]].KEY[inmd->Size];
      if (isList) c = (n%10)+'0';
             else c = (c==0) ? ' ':c;
      sprintf(buff,"%c%s ",c,inmd->KTW.KTW[inmd->KTW.MATCH[id]].WORD);
      strcat(RECORD->ListArea,buff);
    }
    id++; n++;
  } while (id<inmd->KTW.Matches&&len<RECORD->MaxListArea);
  if (id!=inmd->KTW.Matches) {
    strcat(RECORD->ListArea,">");
    inmd->KTW.NextAvail = id;
  } else {
    if (inmd->KTW.Now==0) inmd->KTW.NextAvail = -1; else {
      inmd->KTW.NextAvail =  0;
      strcat(RECORD->ListArea,">");
    }
  }
  sprintf(buff,":%d",inmd->KTW.Matches-inmd->KTW.Now);
  strcat(RECORD->ListArea,buff);
  return 1;
}

int imeFeedKey(iomeRecord *RECORD,int ch,int state)
{
  INMD *inmd = (INMD*)RECORD->Owner;
  int key = iomeTOLOWER(ch);
  int id;
  char * keyname;

  switch (key) {
    case IME_KEY_ESC: /* Clean */
      imePrepare(RECORD);
    return IME_RT_BLOCKOUT;

    case IME_KEY_BACK:   /* Delete */
    case IME_KEY_DELETE:
      if (inmd->Size==0) return IME_RT_REJECT;
      inmd->Size--;
      inmd->KEY[inmd->Size]=0;
      RECORD->KeyArea[inmd->Size*2]=0;
      if (inmd->Size==0) {
        RECORD->ListArea[0]=0;
        inmd->Nomatch = 0;
        inmd->KTW.SameKey = 0;
        return IME_RT_BLOCKOUT;
      }
      if (BuildMatchList(&inmd->KTW,inmd->KEY,inmd->Size)) {
        ShowListArea(RECORD,inmd,inmd->KTW.SameKey);
        inmd->Nomatch = 0;
      } else {
        sprintf(RECORD->ListArea,"No match");
        inmd->Nomatch = 1;
      }
    return IME_RT_BLOCKOUT;

    case '<': /* Previous Page */
      
    break;

    case '>': /* Next Page */
      if (inmd->Nomatch) break;
      if (inmd->Size==0) break;
      if (inmd->KTW.Matches==0) break;
      if (inmd->KTW.NextAvail<0) break;
      inmd->KTW.Now = inmd->KTW.NextAvail;
      ShowListArea(RECORD,inmd,inmd->KTW.SameKey);
    return IME_RT_BLOCKOUT;

    case ' ': /* Select Current precisely */
      if (inmd->Size==0) break;
      if (inmd->KTW.Matches==0) break;
      if (inmd->KTW.SameKey==1&&inmd->KTW.Matches>1) 
        return IME_RT_BLOCKOUT|IME_RT_CRY;
      inmd->KEY[inmd->Size] = 0;
      inmd->Size++;
      if (BuildMatchList(&inmd->KTW,inmd->KEY,inmd->Size)) {
        if (inmd->KTW.SameKey==1) {
          if (inmd->KTW.Matches==1) {
            strcpy(RECORD->Value,inmd->KTW.KTW[inmd->KTW.MATCH[0]].WORD);
            strcpy(RECORD->ListArea,RECORD->Value);
            strcat(RECORD->ListArea,":");
            strcat(RECORD->ListArea,RECORD->KeyArea);
            inmd->Size = 0;
            memset(inmd->KEY,0,64);
            memset(RECORD->KeyArea,0,64);
            return IME_RT_SEND;
          } else
            ShowListArea(RECORD,inmd,1);
        } else {
          ShowListArea(RECORD,inmd,0);
        }
      } else {
        sprintf(RECORD->ListArea,"?");
      }
    break;

    default:
      if (key>=0x100) return IME_RT_REJECT;
      if (key<' '||key>0x7e) {
        return IME_RT_REJECT;
      }
      if (inmd->Nomatch) return IME_RT_BLOCKOUT|IME_RT_CRY;
      keyname = WhichKEY(&inmd->KC,key);
      if (keyname==NULL) IME_RT_REJECT;
      if (inmd->KTW.SameKey==1&&inmd->KTW.Matches>1&&inmd->KEY[inmd->Size]==0) {
        /* Select 0-9 */
        if (key>='0'&&key<='9') {
          key -='0';
          key  = (key == 0) ? 9:key-1;
          key += inmd->KTW.Now;
          if (key>=inmd->KTW.Matches) return IME_RT_BLOCKOUT|IME_RT_CRY;
          strcpy(RECORD->Value,inmd->KTW.KTW[inmd->KTW.MATCH[key]].WORD);
          strcpy(RECORD->ListArea,RECORD->Value);
          strcat(RECORD->ListArea,":");
          strcat(RECORD->ListArea,RECORD->KeyArea);
          inmd->Size = 0;
          inmd->KTW.SameKey = 0;
          memset(inmd->KEY,0,64);
          memset(RECORD->KeyArea,0,64);
          return IME_RT_SEND;
        }
        return IME_RT_BLOCKOUT|IME_RT_CRY;
      }
      inmd->KEY[inmd->Size] = key;
      inmd->Size++;
      strcat(RECORD->KeyArea,keyname);
      if (BuildMatchList(&inmd->KTW,inmd->KEY,inmd->Size)) {
        if (inmd->KTW.SameKey==1) {
          if (inmd->KTW.Matches==1) {
            strcpy(RECORD->Value,inmd->KTW.KTW[inmd->KTW.MATCH[0]].WORD);
            strcpy(RECORD->ListArea,RECORD->Value);
            strcat(RECORD->ListArea,":");
            strcat(RECORD->ListArea,RECORD->KeyArea);
            inmd->Size = 0;
            memset(inmd->KEY,0,64);
            memset(RECORD->KeyArea,0,64);
            return IME_RT_SEND;
          } else
            ShowListArea(RECORD,inmd,1);
        } else ShowListArea(RECORD,inmd,0);
      } else {
        sprintf(RECORD->ListArea,"No match");
        inmd->Nomatch = 1;
      }
  }
  return IME_RT_BLOCKOUT;
}

int imeListArea(iomeRecord *RECORD,int x,int y)
{
  return 0;
}

int imeKeyArea(iomeRecord *RECORD,int x,int y)
{
  return 0;
}

int imeFocus(iomeRecord *RECORD)
{
  return 0;
}

int imeQuit(iomeRecord *RECORD)
{
  INMD *inmd = RECORD->Owner;
  int n;
  
  free(inmd);
  free(RECORD->ListArea);
  free(RECORD->KeyArea);
  free(RECORD->Value);
  return 1;  
}
